use std::{
    ffi::c_void,
    sync::{
        atomic::{AtomicBool, Ordering},
        OnceLock,
    },
    thread::{self, JoinHandle},
    time::Duration,
};

use anyhow::{anyhow, Result};
use chuniio_tasoller_common::{create_input_shared_memory, log::init_logger};
use log::{error, info, warn};
use rusb::{DeviceHandle, GlobalContext};
use winapi::{
    shared::{
        minwindef::{BOOL, DWORD, HINSTANCE, LPVOID, TRUE},
        ntdef::HRESULT,
        winerror::{E_FAIL, S_OK},
    },
    um::winnt::DLL_PROCESS_ATTACH,
};

type SliderCallbackFn = unsafe extern "C" fn(data: *const u8);

static TASOLLER: OnceLock<DeviceHandle<GlobalContext>> = OnceLock::new();
static mut SLIDER_THREAD: OnceLock<JoinHandle<()>> = OnceLock::new();
static SLIDER_ACTIVE: AtomicBool = AtomicBool::new(false);
static mut USB_OUT: [u8; 240] = [0; 240];

#[no_mangle]
extern "system" fn DllMain(_dll_module: HINSTANCE, call_reason: DWORD, _reserved: LPVOID) -> BOOL {
    match call_reason {
        DLL_PROCESS_ATTACH => {
            init_logger();
        }
        _ => {}
    }

    TRUE
}

#[no_mangle]
pub extern "C" fn chuni_io_get_api_version() -> u16 {
    0x0102
}

#[no_mangle]
pub extern "C" fn chuni_io_slider_init() -> HRESULT {
    info!("chuni_io_slider_init");

    match tasoller_init() {
        Ok(_) => S_OK,
        Err(e) => {
            error!("Failed to initialize Tasoller: {e:#}");
            E_FAIL
        }
    }
}

#[no_mangle]
pub extern "C" fn chuni_io_slider_start(callback: *const c_void) {
    if callback.is_null() {
        return;
    }

    unsafe {
        if SLIDER_THREAD.get().is_some() {
            return;
        }

        SLIDER_ACTIVE.store(true, Ordering::Relaxed);

        let callback = std::mem::transmute::<_, SliderCallbackFn>(callback);
        let thread = thread::spawn(move || {
            let input_shmem = match create_input_shared_memory(false) {
                Ok(s) => s,
                Err(e) => {
                    error!("Could not obtain shared memory for Tasoller input: {e:#}");
                    return;
                }
            };
            let usb_in = input_shmem.as_slice();
            let mut pressure = [0u8; 32];

            while SLIDER_ACTIVE.load(Ordering::Relaxed) {
                for i in 0..32 {
                    pressure[if i % 2 == 0 { 30 - i } else { 32 - i }] = usb_in[i + 4];
                }

                callback(pressure.as_ptr());
                thread::sleep(Duration::from_nanos(1_000_000));
            }
        });

        SLIDER_THREAD.set(thread).unwrap();
    }
}

#[no_mangle]
pub extern "C" fn chuni_io_slider_set_leds(data: *const u8) {
    if data.is_null() {
        return;
    }

    unsafe {
        let Some(device) = TASOLLER.get() else { return };
        let led_report = std::slice::from_raw_parts(data, 96);

        for n in 0..31 {
            USB_OUT[n * 3 + 3] = led_report[n * 3 + 2];
            USB_OUT[n * 3 + 4] = led_report[n * 3 + 1];
            USB_OUT[n * 3 + 5] = led_report[n * 3 + 0];
        }

        if let Err(e) = device.write_bulk(0x03, &USB_OUT, Duration::from_micros(1)) {
            error!("Error writing slider LED data to Tasoller: {e:#}");
        }
    }
}

#[no_mangle]
pub unsafe extern "C" fn chuni_io_slider_stop() {
    let Some(thread) = SLIDER_THREAD.take() else {
        return;
    };

    SLIDER_ACTIVE.store(false, Ordering::Relaxed);

    thread.join().expect("Couldn't join slider input thread");
}

#[no_mangle]
pub extern "C" fn chuni_io_led_init() -> HRESULT {
    S_OK
}

#[no_mangle]
pub extern "C" fn chuni_io_led_set_colors(board: u8, data: *const u8) {
    if data.is_null() {
        return;
    }

    let start = if board == 0 { 96 } else { 168 };
    let rgb_start = if board == 0 { 0x96 } else { 0xb4 };

    unsafe {
        let Some(device) = TASOLLER.get() else { return };

        for n in 0..24 {
            USB_OUT[start + n * 3 + 1] = *data.wrapping_add(rgb_start);
            USB_OUT[start + n * 3 + 0] = *data.wrapping_add(rgb_start + 1);
            USB_OUT[start + n * 3 + 2] = *data.wrapping_add(rgb_start + 2);
        }

        if let Err(e) = device.write_bulk(0x03, &USB_OUT, Duration::from_micros(1)) {
            error!("Error writing LED data to Tasoller: {e:#}");
        }
    }
}

fn tasoller_init() -> Result<()> {
    let mut device = match rusb::open_device_with_vid_pid(0x1CCF, 0x2333) {
        Some(dev) => dev,
        None => {
            return Err(anyhow!("Tasoller not found."));
        }
    };

    device
        .claim_interface(0)
        .map_err(|e| anyhow!("Cannot open Tasoller: {e:#}"))?;

    TASOLLER
        .set(device)
        .map_err(|e| anyhow!("Cannot store device handle: {e:#?}"))?;

    unsafe {
        USB_OUT[0] = 0x42;
        USB_OUT[1] = 0x4C;
        USB_OUT[2] = 0x00;
    }

    thread::spawn(input_thread_proc);

    Ok(())
}

fn input_thread_proc() {
    info!("Input thread started");

    let mut shmem = match create_input_shared_memory(true) {
        Ok(s) => s,
        Err(e) => {
            error!("Could not obtain shared memory for Tasoller input: {e:#}");
            return;
        }
    };
    let usb_in = unsafe { shmem.as_slice_mut() };
    let device = TASOLLER.get().unwrap();

    loop {
        if let Err(e) = device.read_interrupt(0x84, usb_in, Duration::from_micros(1)) {
            error!("Failed to read data from Tasoller: {e:#}");
        }

        if usb_in[0] != 0x68 || usb_in[1] != 0x66 || usb_in[2] != 0x84 {
            warn!("Received packet does not start with magic bytes 68 66 84");
        }
    }
}

// Stubs. These are handled by the amdaemon dll.
#[no_mangle]
pub extern "C" fn chuni_io_jvs_init() -> HRESULT {
    S_OK
}

#[no_mangle]
pub extern "C" fn chuni_io_jvs_poll(_opbtn: *mut u8, _beams: *mut u8) {}

#[no_mangle]
pub extern "C" fn chuni_io_jvs_read_coin_counter(_total: *mut u8) {}
